import json
from typing import TYPE_CHECKING, Optional, Union

from .apraw_base import aPRAWBase
from ..enums.distinguishment_option import DistinguishmentOption
from ..subreddit.removal_reasons import SubredditRemovalReason
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit


class ItemModeration:
    """
    A helper class to moderate comments, submissions and modmail.
    """

    def __init__(self, reddit: 'Reddit', item: aPRAWBase):
        """
        Create an instance of ``ItemModeration``.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        item : aPRAWBase
            The item this helper performs requests for.
        """
        self._reddit = reddit
        self._item = item

    async def approve(self):
        """
        Approve the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_approve"], id=self.fullname)

    def _add_removal_reason(self, mod_note: Optional[str] = "", reason: Union[str, SubredditRemovalReason] = None):
        """
        Add a removal reason to a comment or submission.

        It is necessary to first call :meth:`~.remove` on the :class:`~.Comment` or :class:`~.Submission`.
        If ``reason_id`` is not specified, ``mod_note`` cannot be blank.

        Parameters
        ----------
        mod_note: Optional[str]
            A message for the other moderators.
        reason:  str or SubredditRemovalReason
            The removal reason ID or a subreddit removal reason to add.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        if not reason and not mod_note:
            raise ValueError("mod_note cannot be blank if reason_id is not specified")
        # Only the first element of the item_id list is used.
        data = {
            "item_ids": [self.fullname],
            "mod_note": mod_note,
            "reason_id": str(reason),
        }
        return self._reddit.post(API_PATH["removal_reasons"], data={"json": json.dumps(data)})

    async def remove(self, spam: bool = False, mod_note: Optional[str] = "",
                     reason: Union[str, SubredditRemovalReason] = None):
        """
        Remove the Reddit item.

        Parameters
        ----------
        spam: bool
            When ``True``, use the removal to help train the Subreddit's spam filter (default: ``False``).
        mod_note: Optional[str]
            A message for the other moderators.
        reason:  str or SubredditRemovalReason
            The removal reason ID or a subreddit removal reason to add.

        Returns
        -------
        resp: Dict or Tuple
            The API response JSON or a tuple of dictionaries if a removal reason / mod note was added as well.
        """
        data = {"id": self.fullname, "spam": bool(spam)}
        res = await self._reddit.post(API_PATH["mod_remove"], data=data)

        if any([reason, mod_note]):
            res1 = self._add_removal_reason(mod_note, reason)
            return res, res1

        return res

    async def distinguish(self, how: DistinguishmentOption = "yes", sticky: bool = False):
        """
        Distinguish the Reddit item.

        Parameters
        ----------
        how : DistinguishmentOption
            The type of distinguishment to be added to the item.
        sticky : bool, optional
            Whether the item should be stickied.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_distinguish"], id=self.fullname, how=how, sticky=sticky)

    async def undistinguish(self):
        """
        Undistinguish the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_distinguish"], id=self.fullname, how="no", sticky=False)

    async def ignore_reports(self):
        """
        Ignore reports on the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_ignore_reports"], id=self.fullname)

    async def unignore_reports(self):
        """
        Unignore previously ignored reports on the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_unignore_reports"], id=self.fullname)

    @property
    def fullname(self) -> str:
        """
        Retrieve the fullname of the item this helper performs requests for.

        Returns
        -------
        fullname: str
            The ID prepended with the kind of the item this helper belongs to.
        """
        return self._item.fullname


class PostModeration(ItemModeration):
    """
    A helper class to moderate comments and submissions.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    fullname: str
        The ID prepended with the kind of the item this helper belongs to.
    """

    async def lock(self):
        """
        Lock the item from further replies.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_lock"], id=self.fullname)

    async def unlock(self):
        """
        Unlock the item from further replies.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post(API_PATH["mod_unlock"], id=self.fullname)



from typing import Dict, Any, TYPE_CHECKING

from ..helpers.apraw_base import aPRAWBase
from ..mixins.author import AuthorMixin
from ..mixins.replyable import ReplyableMixin
from ..mixins.subreddit import SubredditMixin

if TYPE_CHECKING:
    from ...reddit import Reddit


class Message(aPRAWBase, SubredditMixin, AuthorMixin, ReplyableMixin):
    """
    The model representing comments.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    =========================== ======================================================================================
    Attribute                   Description
    =========================== ======================================================================================
    ``first_message``           The first message sent in the message thread if the current message wasn't the first.
    ``first_message_name``      The fullname of the first message in the message thread if applicable.
    ``subreddit``               The subreddit this conversation is being held in if applicable.
    ``likes``                   ``None``
    ``replies``                 A list of all the message replies if applicable, otherwise an empty string.
    ``id``                      The message ID.
    ``subject``                 The subject of this message's thread.
    ``associated_awarding_id``  The ID of the associated awarding if the message was sent in the context of an award.
    ``score``                   ``0``
    ``author``                  The username of the message's author.
    ``num_comments``            The number of comments in this message's thread.
    ``parent_id``               ``None``
    ``subreddit_name_prefixed`` The prefixed name of the subreddit this conversation is being held in if applicable.
    ``new``                     ``bool``
    ``type``                    ``str``
    ``body``                    The markdown string contents of this message.
    ``dest``                    The recipient of the message.
    ``body_html``               The HTML string contents of this message.
    ``was_comment``             Whether this message was a comment.
    ``name``                    The fullname of this message, representing the ID prefixed with its kind. (e.g. `t4_`)
    ``created``                 The timestamp on which this message was created.
    ``created_utc``             The parsed UTC ``datetime`` on which this message was created.
    ``context``                 ``str``
    ``distinguished``           The type of distinguishment on this message object.
    =========================== ======================================================================================

    """

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any]):
        """
        Create an instance of a Message object.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data, reddit.message_kind)
        AuthorMixin.__init__(self)
        SubredditMixin.__init__(self)



from typing import TYPE_CHECKING, Dict, Any

from ..helpers.apraw_base import aPRAWBase
from ..mixins.subreddit import SubredditMixin
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .subreddit import Subreddit


class SubredditSettings(aPRAWBase, SubredditMixin):
    """
    A model representing subreddit settings.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================================= =============================================================================
    Attribute                         Description
    ================================= =============================================================================
    ``all_original_content``          Whether the subreddit only allows original content.
    ``allow_chat_post_creation``      Whether the subreddit allows chat post creation.
    ``allow_discovery``               Whether this subreddit can be discovered through the recommendations.
    ``allow_galleries``               Whether this subreddit allows submissions with galleries.
    ``allow_images``                  Whether this subreddit allows image posts.
    ``allow_polls``                   Whether this subreddit allows poll posts.
    ``allow_post_crossposts``         Whether this subreddit allows crossposts.
    ``allow_videos``                  Whether this subreddit allows video submissions.
    ``collapse_deleted_comments``     Whether deleted comments in threads should be automatically collapsed.
    ``comment_score_hide_mins``       The comment score below which comments should be hidden.
    ``content_options``               ``string``
    ``crowd_control_chat_level``      ``int``
    ``crowd_control_level``           ``int``
    ``crowd_control_mode``            ``bool``
    ``default_set``                   ``bool``
    ``description``                   The subreddit's short description.
    ``disable_contributor_requests``  ``bool``
    ``domain``                        ``None``
    ``exclude_banned_modqueue``       Whether banned users should be excluded from the modqueue.
    ``free_form_reports``             Whether users can submit custom text reports.
    ``header_hover_text``             The hover text for the subreddit's header.
    ``hide_ads``                      Whether ads should be hidden on this subreddit.
    ``key_color``                     ``string``
    ``language``                      The subreddit's default language as a language code (i.e. "en" for English).
    ``original_content_tag_enabled``  Whether the subreddit has the OC tag enabled.
    ``over_18``                       Whether this subreddit is marked NSFW.
    ``public_description``            The subreddit's public description.
    ``public_traffic``                ``bool``
    ``restrict_commenting``           Whether comments are restricted on the subreddit.
    ``restrict_posting``              Whether posts are restricted on the subreddit.
    ``show_media_preview``            Whether media previews should be displayed by clients.
    ``show_media``                    ``bool``
    ``spam_comments``                 The comment spam filter's setting, either "low", "medium" or "high".
    ``spam_links``                    The link spam filter's setting, either "low", "medium" or "high".
    ``spam_selfposts``                The selfpost spam filter's setting, either "low", "medium" or "high".
    ``spoilers_enabled``              Whether the spoiler marker has been enabled on this subreddit.
    ``submit_link_label``             The submit button's label.
    ``submit_text_label``             The submit text's label.
    ``submit_text``                   ``string``
    ``subreddit_id``                  The ID of the subreddit with the prepended kind i.e. ``t5_``.
    ``subreddit_type``                One of "public", "private" or "restricted".
    ``suggested_comment_sort``        The default comment sort for submissions.
    ``title``                         The subreddit's name.
    ``toxicity_threshold_chat_level`` ``int``
    ``welcome_message_enabled``       Whether the subreddit has enabled welcome messages.
    ``welcome_message_text``          The welcome message's text of this subreddit.
    ``wiki_edit_age``                 The minimum account age requirement for wiki editors.
    ``wiki_edit_karma``               The minimum account karma requirement for wiki editors.
    ``wikimode``                      The mode the wiki is in e.g. "modonly".
    ================================= =============================================================================

   """

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any], subreddit: 'Subreddit' = None):
        """
        Create an instance of ``SubredditSettings``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        subreddit: Subreddit
            The subreddit these settings belong to.
        """
        data.update({"subreddit": data.get("subreddit") or data.get("title")})
        aPRAWBase.__init__(self, reddit, data, reddit.subreddit_settings_kind)
        SubredditMixin.__init__(self, subreddit)

    async def fetch(self) -> 'SubredditSettings':
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: SubredditSettings
            The ``SubredditSettings`` model with updated data.
        """
        sub = self._data.get("subreddit") or self._data.get("title")
        resp = await self._reddit.get(API_PATH["subreddit_settings"].format(sub=sub))
        self._update(resp["data"])
        return self



import asyncio
from datetime import datetime, timedelta
from functools import wraps
from typing import Dict, Callable, Any, Awaitable, Optional

from multidict import CIMultiDictProxy

from .const import BASE_URL
from .models import User


class RequestHandler:

    def __init__(self, user: User):
        self.user = user
        self.queue = []

    async def get_request_headers(self) -> Dict:
        if self.user.token_expires <= datetime.now():
            url = "https://www.reddit.com/api/v1/access_token"
            session = await self.user.auth_session()

            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": self.user.user_agent
            }

            resp = await session.post(url, data=self.user.password_grant, headers=headers)

            async with resp:
                if resp.status == 200:
                    self.user.access_data = await resp.json()
                    self.user.token_expires = datetime.now(
                    ) + timedelta(seconds=self.user.access_data["expires_in"])
                else:
                    raise Exception("Invalid user data.")

        return {
            "Authorization": "{} {}".format(self.user.access_data["token_type"], self.user.access_data["access_token"]),
            "User-Agent": self.user.user_agent
        }

    def update(self, data: CIMultiDictProxy):
        if "x-ratelimit-remaining" in data:
            self.user.ratelimit_remaining = int(float(data["x-ratelimit-remaining"]))
        if "x-ratelimit-used" in data:
            self.user.ratelimit_used = int(data["x-ratelimit-used"])
        if "x-ratelimit-reset" in data:
            self.user.ratelimit_reset = datetime.now() + timedelta(seconds=int(data["x-ratelimit-reset"]))

    async def close(self):
        await self.user.close()

    class Decorators:

        @classmethod
        def check_ratelimit(
                cls, func: Callable[[Any, Any], Awaitable[Any]]) -> Callable[[Any, Any], Awaitable[Any]]:
            @wraps(func)
            async def execute_request(self, *args, **kwargs) -> Any:
                id = datetime.now().strftime('%Y%m%d%H%M%S')
                self.queue.append(id)

                if self.user.ratelimit_remaining < 1:
                    execution_time = self.user.ratelimit_reset + \
                                     timedelta(seconds=len(self.queue))
                    wait_time = (execution_time - datetime.now()).total_seconds()
                    await asyncio.sleep(wait_time)

                result = await func(self, *args, **kwargs)
                self.queue.remove(id)
                return result

            return execute_request

    @Decorators.check_ratelimit
    async def get(self, endpoint: Optional[str] = "", _url: Optional[str] = "", **kwargs) -> Any:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        if endpoint:
            url = BASE_URL.format(endpoint, "&".join(params))
        elif _url:
            url = _url + "?" + "&".join(params)
        else:
            raise ValueError("One of endpoint or _url must be specified.")

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.get(url, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()

    @Decorators.check_ratelimit
    async def delete(self, endpoint: str = "", **kwargs) -> Any:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        url = BASE_URL.format(endpoint, "&".join(params))

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.delete(url, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()

    @Decorators.check_ratelimit
    async def put(self, endpoint: str = "", data: Dict = None, **kwargs) -> Any:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        url = BASE_URL.format(endpoint, "&".join(params))

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.delete(url, data=data, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()

    @Decorators.check_ratelimit
    async def post(self, endpoint: str = "", url: str = "", data: Dict = None, **kwargs) -> Any:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        if endpoint:
            url = BASE_URL.format(endpoint, "&".join(params))
        elif url:
            url = "{}?{}".format(url, "&".join(params))

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.post(url, data=data, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()



from datetime import datetime
from typing import TYPE_CHECKING, Union, Optional, Dict

from ..helpers.apraw_base import aPRAWBase
from ..helpers.generator import ListingGenerator
from ..helpers.streamable import streamable
from ..mixins.redditor import RedditorMixin
from ..reddit.listing import Listing
from ..reddit.redditor import Redditor
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .subreddit import Subreddit


class BannedUser(aPRAWBase, RedditorMixin):
    """
    The model representing banned users on a subreddit. The Redditor can be retrieved via :meth:`~apraw.models.BannedUser.redditor()`.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ========================== ========================================================
    Attribute                  Description
    ========================== ========================================================
    ``banned``                 The parsed UTC date on which the user was banned.
    ``date``                   The UTC timestamp on which the user was banned.
    ``days_left``              The number of days left for the ban. ``0`` if permanent.
    ``id``                     The Redditor's fullname (t2_ID).
    ``name``                   The Redditor's name.
    ``note``                   The ban note added by the subreddit moderators.
    ``rel_id``                 ``str``
    ========================== ========================================================

    """

    def __init__(self, reddit: 'Reddit', data: Dict, subreddit: 'Subreddit'):
        """
        Create a BannedUser instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the API.
        """
        if "date" in data:
            data["banned"] = datetime.utcfromtimestamp(data["date"])
        super().__init__(reddit, data, reddit.account_kind)
        self._subreddit = subreddit


class BannedListing(Listing):
    """
    A model representing listings of banned users.
    """

    def __getitem__(self, index: int) -> BannedUser:
        """
        Return the item at position index in the list.

        Parameters
        ----------
        index : int
            The item's index.

        Returns
        -------
        item: aPRAWBase
            The searched item.
        """
        item = getattr(self, self.CHILD_ATTRIBUTE)[index]

        if isinstance(item, aPRAWBase):
            return item
        else:
            return BannedUser(self._reddit, item, self._subreddit)


class SubredditBanned:
    """
    A helper class for interacting with a subreddit's banned users.

    """

    def __init__(self, reddit: 'Reddit', subreddit: 'Subreddit'):
        """
        Create an instance of ``SubredditBanned``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        subreddit: Subreddit
            The subreddit this helper instance belongs to and performs requests for.
        """
        self._reddit = reddit
        self._subreddit = subreddit

    #: Streamable listing endpoint.
    @streamable
    async def __call__(self, redditor: Optional[Union[str, Redditor]] = None, limit: int = 100, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the subreddit's banned users.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.banned.stream():
                    print(comment)

        Parameters
        ----------
        redditor: Redditor or str
            A single Redditor to search for, useful for checking if they have already been banned.
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the subreddit's banned users.
        """
        url = API_PATH["subreddit_banned"].format(sub=self._subreddit.display_name)
        if redditor:
            kwargs["redditor"] = str(redditor)
        return ListingGenerator(self._reddit, url, limit, self._subreddit, listing_class=BannedListing, **kwargs)

    async def add(self, redditor: Union[str, Redditor], **kwargs):
        r"""
        Ban a Reddit in the respective subreddit.

        Parameters
        ----------
        redditor: Redditor or str
            The Redditor to be banned.
        duration: int
            An optional duration if the ban is not permanent.
        ban_context: str
            Fullname of a thing to be used as the context.
        ban_message: str
            The message to be sent to the banned Redditor.
        ban_reason: str
            An additional reason for the ban.
        note: str
            An internal mod note.

        Returns
        -------
        resp: Dict
            The raw response JSON dictionary.
        """
        url = API_PATH["sub_friend"].format(sub=self._subreddit.display_name)
        data = {"name": str(redditor), "type": "banned"}
        data.update(**kwargs)
        return await self._reddit.post(url, data=data)

    async def remove(self, redditor: Union[str, Redditor]):
        """
        Unban a Reddit in the respective subreddit.

        Parameters
        ----------
        redditor: Redditor or str
            The Redditor to be banned.

        Returns
        -------
        resp: Dict
            The raw response JSON dictionary.
        """
        url = API_PATH["sub_unfriend"].format(sub=self._subreddit.display_name)
        return await self._reddit.post(url, data={"name": str(redditor), "type": "banned"})



from typing import Any


class BoundedSet:
    """
    A set with a maximum size that evicts the oldest items when necessary.
    This class does not implement the complete set interface.
    """

    def __init__(self, max_items: int):
        """Construct an instance of the BoundedSet."""
        self.max_items = max_items
        self._list = list()
        self._set = set()

    def __contains__(self, item: Any) -> bool:
        """Test if the BoundedSet contains item."""
        return item in self._set

    def add(self, item: Any):
        """Add an item to the set discarding the oldest item if necessary."""
        if len(self._set) == self.max_items:
            self._set.remove(self._list.pop(0))
        self._list.append(item)
        self._set.add(item)



from ..reddit.redditor import Redditor


class RedditorMixin:
    """
    Mixin for items that represent a slimmed version of a Redditor.
    """

    def __str__(self):
        """
        Returns the Redditor's name.

        Returns
        -------
        name: str
            The Redditor's name.
        """
        return getattr(self, "name")

    async def redditor(self) -> Redditor:
        """
        Retrieve the Redditor this Moderator represents.

        Returns
        -------
        redditor: Redditor
            The Redditor that is represented by this object.
        """
        return await self._reddit.redditor(getattr(self, "name"))
